﻿#include "SoundPreviewer.h"
#include "cri_xpt.h"
#include "cri_error.h"
#ifdef XPT_TGT_PC
#include "cri_atom_wasapi.h"
#endif
#ifdef XPT_TGT_MACOSX
#include <unistd.h>
#include "cri_atom_macosx.h"
#define memcpy_s(a, b, c, d)    memcpy((a), (c), (d))
#define _stricmp strcasecmp
#endif
#define	UNUSED(a)				{if((a)==(a)){}}
#define SERVERFREQUENCY     (150)

namespace {
    const int avoidMultiStartMs = 100;
    const int statusCheckMs = 16;
}

static void user_error_callback_func(
    const CriChar8 *errid, CriUint32 p1, CriUint32 p2, CriUint32 *parray)
{
    const CriChar8 *errmsg;
    UNUSED(parray);

    /* エラー文字列の表示 */
    errmsg = criErr_ConvertIdToMessage(errid, p1, p2);
    printf("%s\n", errmsg);
    fflush(stdout);
}

static void *user_alloc_func(void *obj, CriUint32 size)
{
    void *ptr;
    UNUSED(obj);
    ptr = malloc(size);
    return (ptr);
}

static void user_free_func(void *obj, void *ptr)
{
    UNUSED(obj);
    free(ptr);
}

SoundPreviewer::SoundPreviewer(QObject *parent)
    : QObject(parent)
    , preciseTimer(new QTimer(this))
    , preciseFadeOutTimer(new QTimer(this))
    , avoidMultiStartTimer(new QTimer(this))
    , statusCheckTimer(new QTimer(this))
    , previousPlaybackPosIsZero(true)
    , fadeoutMsReserved(0)
{
    this->preciseTimer->setTimerType(Qt::PreciseTimer);
    this->preciseFadeOutTimer->setTimerType(Qt::PreciseTimer);
    this->avoidMultiStartTimer->setInterval(::avoidMultiStartMs);
    this->avoidMultiStartTimer->setSingleShot(true);
    this->statusCheckTimer->setInterval(::statusCheckMs);
    connect(this->statusCheckTimer, &QTimer::timeout, this, &SoundPreviewer::statusCheckTimeout);
    this->statusCheckTimer->start();
    connect(this->preciseTimer, &QTimer::timeout, this, &SoundPreviewer::preciseTimerTimeout);
    connect(this->preciseFadeOutTimer, &QTimer::timeout, this, &SoundPreviewer::preciseFadeOutTimerTimeout);
    /* エラーコールバック関数の登録 */
    criErr_SetCallback(user_error_callback_func);
    /* メモリアロケータの登録 */
    criAtomEx_SetUserAllocator(user_alloc_func, user_free_func, NULL);

    /* ライブラリの初期化 */
#ifdef XPT_TGT_PC
    {
        CriAtomExConfig_WASAPI config_wasapi;
        criAtomEx_SetDefaultConfig_WASAPI(&config_wasapi);
        config_wasapi.atom_ex.server_frequency = SERVERFREQUENCY;
        /* 各サブモジュールのサーバ動作周波数をあわせる */
        config_wasapi.asr.server_frequency		= config_wasapi.atom_ex.server_frequency;
        config_wasapi.hca_mx.server_frequency	= config_wasapi.atom_ex.server_frequency;
        criAtomEx_Initialize_WASAPI(&config_wasapi, NULL, 0);
    }
#elif defined XPT_TGT_ANDROID
    {
        CriAtomExConfig_ANDROID config_android;
        criAtomEx_SetDefaultConfig_ANDROID(&config_android);
        /* 各サブモジュールのサーバ動作周波数をあわせる */
        config_android.atom_ex.server_frequency = SERVERFREQUENCY;
        config_android.asr.server_frequency		= config_android.atom_ex.server_frequency;
        config_android.hca_mx.server_frequency	= config_android.atom_ex.server_frequency;
        criAtomEx_Initialize_ANDROID(&config_android, NULL, 0);
    }
#else
    {
        CriAtomExConfig_MACOSX config_macosx;
        criAtomEx_SetDefaultConfig_MACOSX(&config_macosx);
        /* 各サブモジュールのサーバ動作周波数をあわせる */
        config_macosx.atom_ex.server_frequency = SERVERFREQUENCY;
        config_macosx.asr.server_frequency		= config_macosx.atom_ex.server_frequency;
        config_macosx.hca_mx.server_frequency	= config_macosx.atom_ex.server_frequency;
        criAtomEx_Initialize_MacOSX(&config_macosx, NULL, 0);
    }
#endif

    /* D-basの作成 */
    this->dbas_id = criAtomDbas_Create(NULL, NULL, 0);

    /*	wavファイル再生用ボイスプールの設定	*/
    {
        CriAtomExWaveVoicePoolConfig voice_pool_config;

        criAtomExVoicePool_SetDefaultConfigForWaveVoicePool(&voice_pool_config);
        voice_pool_config.player_config.streaming_flag = CRI_TRUE;
        voice_pool_config.num_voices = 16;
        voice_pool_config.player_config.max_channels = 8;
        voice_pool_config.player_config.max_sampling_rate = 192000;
        this->pool_wave = criAtomExVoicePool_AllocateWaveVoicePool(&voice_pool_config, NULL, 0);
    }
    /* プレイヤーの作成 */
    {
        CriAtomExPlayerConfig player_config;
        criAtomExPlayer_SetDefaultConfig(&player_config);
        player_config.max_path_strings = 2048;
        player_config.max_path = 2048;
        this->player = criAtomExPlayer_Create(&player_config, NULL, 0);
    }
    /* WAVE用に指定 */
    criAtomExPlayer_SetFormat(this->player, CRIATOMEX_FORMAT_WAVE);
    /* フェードイン用にFaderを登録する */
    criAtomExPlayer_AttachFader(this->player, NULL, NULL, 0);
    /* フェードアウトに使用するTweenを作成する */
    {
        CriAtomExTweenConfig fadeout_config;
        criAtomExTween_SetDefaultConfig(&fadeout_config);
        fadeout_config.parameter_type = CRIATOMEX_PARAMETER_TYPE_BASIC;
        fadeout_config.id.parameter_id = CRIATOMEX_PARAMETER_ID_VOLUME;
        this->fadeout_tween = criAtomExTween_Create(&fadeout_config, NULL, 0);
    }
}

SoundPreviewer::~SoundPreviewer()
{
    /* プレーヤハンドルの破棄 */
    criAtomExPlayer_Destroy(this->player);
    /* ボイスプールの破棄 */
    criAtomExVoicePool_Free(this->pool_wave);
    /* D-Basの破棄 */
    criAtomDbas_Destroy(this->dbas_id);
    /* ライブラリの終了 */
#ifdef XPT_TGT_PC
    criAtomEx_Finalize_WASAPI();
#elif defined XPT_TGT_ANDROID
    criAtomEx_Finalize_ANDROID();
#else
    criAtomEx_Finalize_MacOSX();
#endif
}

//void SoundPreviewer::PreviewStart(
//        const std::string &filepath,
//        int startTimeMs,
//        int durationMs)
//{
//    int fadeinMs = 0;
//    int fadeoutMs = 0;
//    if(this->avoidMultiStartTimer->isActive()){
//        return;
//    }
//    /* 再生中ならば停止する */
//    CriAtomExPlayerStatus status = criAtomExPlayer_GetStatus(this->player);
//    /* 停止の完了を待つ */
//    if ( status != CRIATOMEXPLAYER_STATUS_STOP ) {
//        criAtomExPlayer_StopWithoutReleaseTime(this->player);
//        for (;;) {
//            status = criAtomExPlayer_GetStatus(this->player);
//            if ( status == CRIATOMEXPLAYER_STATUS_STOP || status == CRIATOMEXPLAYER_STATUS_ERROR) {
//                break;
//            }
//        }
//    }
//    this->preciseTimer->stop();
//    this->preciseFadeOutTimer->stop();
//    criAtomExPlayer_DetachTweenAll(this->player);
//    criAtomExPlayer_SetFile(this->player, NULL, filepath.c_str());
//    criAtomExPlayer_SetStartTime(this->player, (CriSint64)startTimeMs);
//    /* フェードイン設定 */
//    criAtomExPlayer_SetFadeInTime(this->player, fadeinMs);
//    /* 各種タイマーの実行 */
//    this->preciseTimer->setInterval(durationMs);
//    this->preciseTimer->start();
//    this->avoidMultiStartTimer->start();
//    if(0 < fadeoutMs){
//        criAtomExTween_Reset(this->fadeout_tween);
//        criAtomExPlayer_AttachTween(this->player, this->fadeout_tween);
//        this->fadeoutMsReserved = fadeoutMs;
//        this->preciseFadeOutTimer->setInterval(qMax(0, durationMs - fadeoutMs));
//        this->preciseFadeOutTimer->start();
//    }
//    /* 再生指示 */
//    criAtomExPlayer_Start(this->player);
//}

void SoundPreviewer::PreviewStart(
        const std::string& filepath,
        int startTimeSample,
        int sampleRate)
{
    int startTimeMs = (int)(startTimeSample / (sampleRate / 1000.0));
    if(this->avoidMultiStartTimer->isActive()){
        return;
    }
    /* 再生中ならば停止する */
    CriAtomExPlayerStatus status = criAtomExPlayer_GetStatus(this->player);
    /* 停止の完了を待つ */
    if ( status != CRIATOMEXPLAYER_STATUS_STOP ) {
        criAtomExPlayer_StopWithoutReleaseTime(this->player);
        for (;;) {
            status = criAtomExPlayer_GetStatus(this->player);
            if ( status == CRIATOMEXPLAYER_STATUS_STOP || status == CRIATOMEXPLAYER_STATUS_ERROR) {
                break;
            }
        }
    }
    this->preciseTimer->stop();
    this->preciseFadeOutTimer->stop();
    criAtomExPlayer_DetachTweenAll(this->player);
    criAtomExPlayer_SetFile(this->player, NULL, filepath.c_str());
    criAtomExPlayer_SetStartTime(this->player, (CriSint64)startTimeMs);
    criAtomExPlayer_SetFadeInTime(this->player, 0);
    /* タイマーは多重再生禁止のみ稼働 */
    this->avoidMultiStartTimer->start();
    /* 再生指示 */
    criAtomExPlayer_Start(this->player);
}

void SoundPreviewer::PreviewStop()
{
    criAtomExPlayer_StopWithoutReleaseTime(this->player);
    /* 停止の完了を待たない */
}

bool SoundPreviewer::IsPlaying()
{
    CriAtomExPlayerStatus status = criAtomExPlayer_GetStatus(this->player);
    return (status == CRIATOMEXPLAYER_STATUS_PLAYING)
            || (status == CRIATOMEXPLAYER_STATUS_PREP);
}

void SoundPreviewer::preciseTimerTimeout()
{
    this->preciseTimer->stop();
    this->PreviewStop();
}

void SoundPreviewer::preciseFadeOutTimerTimeout()
{
    this->preciseFadeOutTimer->stop();
    /* Tweenを使用してフェードアウトを行う */
    criAtomExTween_MoveTo(this->fadeout_tween, this->fadeoutMsReserved, 0.0f);
}

void SoundPreviewer::statusCheckTimeout()
{
    /* 再生時刻位置の更新を通知する */
    int timeMs = 0;
    CriAtomExPlaybackId id = criAtomExPlayer_GetLastPlaybackId(this->player);
    if(id != CRIATOMEX_INVALID_PLAYBACK_ID){
        CriSint64 samples = 0;
        CriSint32 sample_rate = 1;
        CriBool ret = criAtomExPlayback_GetNumPlayedSamples(id, &samples, &sample_rate);
        if(ret != CRI_FALSE){
            timeMs = (int)(samples * (1000.0 / sample_rate));
        }
    }
    if(0 < timeMs){
        this->previousPlaybackPosIsZero = false;
        emit this->UpdatePlaybackPos(timeMs);
    } else {
        if(!this->previousPlaybackPosIsZero){
            emit this->UpdatePlaybackPos(0);
        }
        this->previousPlaybackPosIsZero = true;
    }
}
